#!/usr/bin/env python
# -*- coding: utf-8 -*-

_author__ = "Patrice Guyot"
__email__ = "patrice.guyot@irit.fr"
__version__ = "1.1.0"

"""
***********************************************************************
Name            : yin.py
Description     : Fundamental frequency estimation. Based on the YIN alorgorithm [1]: De Cheveigné, A., & Kawahara, H. (2002). YIN, a fundamental frequency estimator for speech and music. The Journal of the Acoustical Society of America, 111(4), 1917-1930.
Author          : Patrice Guyot. Previous works on the implementation of the YIN algorithm have been made thanks to Robin Larvor, Maxime Le Coz and Lionel Koenig.
***********************************************************************

"""


import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import fftconvolve
from os import sep
from .audio_processing import *
import time



def differenceFunction_original(x, N, tau_max):
    """
    Compute difference function of data x. This corresponds to equation (6) in [1]

    Original algorithm.

    :param x: audio data
    :param N: length of data
    :param tau_max: integration window size
    :return: difference function
    :rtype: list
    """
    df = [0] * tau_max
    for tau in range(1, tau_max):
         for j in range(0, N - tau_max):
             tmp = long(x[j] - x[j + tau])
             df[tau] += tmp * tmp
    return df


def differenceFunction_scipy(x, N, tau_max):
    """
    Compute difference function of data x. This corresponds to equation (6) in [1]

    Faster implementation of the difference function.
    The required calculation can be easily evaluated by Autocorrelation function or similarly by convolution.
    Wiener–Khinchin theorem allows computing the autocorrelation with two Fast Fourier transforms (FFT), with time complexity O(n log n).
    This function use an accellerated convolution function fftconvolve from Scipy package.

    :param x: audio data
    :param N: length of data
    :param tau_max: integration window size
    :return: difference function
    :rtype: list
    """
    x = np.array(x, np.float64)
    w = x.size
    x_cumsum = np.concatenate((np.array([0]), (x * x).cumsum()))
    conv = fftconvolve(x, x[::-1])
    tmp = x_cumsum[w:0:-1] + x_cumsum[w] - x_cumsum[:w] - 2 * conv[w - 1:]
    return tmp[:tau_max + 1]




def differenceFunction(x, N, tau_max):
    """
    Compute difference function of data x. This corresponds to equation (6) in [1]

    Fastest implementation. Use the same approach than differenceFunction_scipy.
    This solution is implemented directly with Numpy fft.


    :param x: audio data
    :param N: length of data
    :param tau_max: integration window size
    :return: difference function
    :rtype: list
    """

    x = np.array(x, np.float64)
    w = x.size
    tau_max = min(tau_max, w)
    x_cumsum = np.concatenate((np.array([0.]), (x * x).cumsum()))
    size = w + tau_max
    p2 = (size // 32).bit_length()
    nice_numbers = (16, 18, 20, 24, 25, 27, 30, 32)
    size_pad = min(x * 2 ** p2 for x in nice_numbers if x * 2 ** p2 >= size)
    fc = np.fft.rfft(x, size_pad)
    conv = np.fft.irfft(fc * fc.conjugate())[:tau_max]
    return x_cumsum[w:w - tau_max:-1] + x_cumsum[w] - x_cumsum[:tau_max] - 2 * conv



def cumulativeMeanNormalizedDifferenceFunction(df, N):
    """
    Compute cumulative mean normalized difference function (CMND).

    This corresponds to equation (8) in [1]

    :param df: Difference function
    :param N: length of data
    :return: cumulative mean normalized difference function
    :rtype: list
    """

    cmndf = df[1:] * range(1, N) / np.cumsum(df[1:]).astype(float) #scipy method
    return np.insert(cmndf, 0, 1)



def getPitch(cmdf, tau_min, tau_max, harmo_th=0.1):
    """
    Return fundamental period of a frame based on CMND function.

    :param cmdf: Cumulative Mean Normalized Difference function
    :param tau_min: minimum period for speech
    :param tau_max: maximum period for speech
    :param harmo_th: harmonicity threshold to determine if it is necessary to compute pitch frequency
    :return: fundamental period if there is values under threshold, 0 otherwise
    :rtype: float
    """
    tau = tau_min
    while tau < tau_max:
        if cmdf[tau] < harmo_th:
            while tau + 1 < tau_max and cmdf[tau + 1] < cmdf[tau]:
                tau += 1
            return tau
        tau += 1

    return 0    # if unvoiced



def compute_yin(sig, sr, dataFileName=None, w_len=512, w_step=256, f0_min=100, f0_max=500, harmo_thresh=0.1):
    """

    Compute the Yin Algorithm. Return fundamental frequency and harmonic rate.

    :param sig: Audio signal (list of float)
    :param sr: sampling rate (int)
    :param w_len: size of the analysis window (samples)
    :param w_step: size of the lag between two consecutives windows (samples)
    :param f0_min: Minimum fundamental frequency that can be detected (hertz)
    :param f0_max: Maximum fundamental frequency that can be detected (hertz)
    :param harmo_tresh: Threshold of detection. The yalgorithmù return the first minimum of the CMND fubction below this treshold.

    :returns:

        * pitches: list of fundamental frequencies,
        * harmonic_rates: list of harmonic rate values for each fundamental frequency value (= confidence value)
        * argmins: minimums of the Cumulative Mean Normalized DifferenceFunction
        * times: list of time of each estimation
    :rtype: tuple
    """

    # print('Yin: compute yin algorithm')
    tau_min = int(sr / f0_max)
    tau_max = int(sr / f0_min)
    
    timeScale = range(0, len(sig) - w_len, w_step)  # time values for each analysis window
    times = [t/float(sr) for t in timeScale]
    frames = [sig[t:t + w_len] for t in timeScale]

    pitches = [0.0] * len(timeScale)
    harmonic_rates = [0.0] * len(timeScale)
    argmins = [0.0] * len(timeScale)

    for i, frame in enumerate(frames):

        #Compute YIN
        df = differenceFunction(frame, w_len, tau_max)
        cmdf = cumulativeMeanNormalizedDifferenceFunction(df, tau_max)
        p = getPitch(cmdf, tau_min, tau_max, harmo_thresh)
        
        #Get results
        if np.argmin(cmdf)>tau_min:
            argmins[i] = float(sr / np.argmin(cmdf))
        if p != 0: # A pitch was found
            pitches[i] = float(sr / p)
            harmonic_rates[i] = cmdf[p]
        else: # No pitch, but we compute a value of the harmonic rate
            harmonic_rates[i] = min(cmdf)


    if dataFileName is not None:
        np.savez(dataFileName, times=times, sr=sr, w_len=w_len, w_step=w_step, f0_min=f0_min, f0_max=f0_max, harmo_thresh=harmo_thresh, pitches=pitches, harmonic_rates=harmonic_rates, argmins=argmins)
        print('\t- Data file written in: ' + dataFileName)

    return pitches, harmonic_rates, argmins, times


def main(audioFileName="whereIam.wav", w_len=1024, w_step=256, f0_min=70, f0_max=200, harmo_thresh=0.85, audioDir="./", dataFileName=None, verbose=4):
    """
    Run the computation of the Yin algorithm on a example file.

    Write the results (pitches, harmonic rates, parameters ) in a numpy file.

    :param audioFileName: name of the audio file
    :type audioFileName: str
    :param w_len: length of the window
    :type wLen: int
    :param wStep: length of the "hop" size
    :type wStep: int
    :param f0_min: minimum f0 in Hertz
    :type f0_min: float
    :param f0_max: maximum f0 in Hertz
    :type f0_max: float
    :param harmo_thresh: harmonic threshold
    :type harmo_thresh: float
    :param audioDir: path of the directory containing the audio file
    :type audioDir: str
    :param dataFileName: file name to output results
    :type dataFileName: str
    :param verbose: Outputs on the console : 0-> nothing, 1-> warning, 2 -> info, 3-> debug(all info), 4 -> plot + all info
    :type verbose: int
    """

    if audioDir is not None:
        audioFilePath = audioDir + sep + audioFileName
    else:
        audioFilePath = audioFileName

    sr, sig = audio_read(audioFilePath, formatsox=False)

    start = time.time()
    pitches, harmonic_rates, argmins, times = compute_yin(sig, sr, dataFileName, w_len, w_step, f0_min, f0_max, harmo_thresh)
    end = time.time()
    print("Yin computed in: ", end - start)

    duration = len(sig)/float(sr)

    if verbose >3:
        ax1 = plt.subplot(4, 1, 1)
        ax1.plot([float(x) * duration / len(sig) for x in range(0, len(sig))], sig)
        ax1.set_title('Audio data')
        ax1.set_ylabel('Amplitude')
        ax2 = plt.subplot(4, 1, 2)
        ax2.plot([float(x) * duration / len(pitches) for x in range(0, len(pitches))], pitches)
        ax2.set_title('F0')
        ax2.set_ylabel('Frequency (Hz)')
        ax3 = plt.subplot(4, 1, 3, sharex=ax2)
        ax3.plot([float(x) * duration / len(harmonic_rates) for x in range(0, len(harmonic_rates))], harmonic_rates)
        ax3.plot([float(x) * duration / len(harmonic_rates) for x in range(0, len(harmonic_rates))], [harmo_thresh] * len(harmonic_rates), 'r')
        ax3.set_title('Harmonic rate')
        ax3.set_ylabel('Rate')
        ax4 = plt.subplot(4, 1, 4, sharex=ax2)
        ax4.plot([float(x) * duration / len(argmins) for x in range(0, len(argmins))], argmins)
        ax4.set_title('Index of minimums of CMND')
        ax4.set_ylabel('Frequency (Hz)')
        ax4.set_xlabel('Time (seconds)')
        plt.show()



if __name__ == '__main__':
    main()





#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
CLEESE toolbox v2.0
jan 2022, Lara Kermarec <lara.git@kermarec.bzh> for CNRS

Image deformation engine for CLEESE, using mediapipe
'''

import numpy as np
from numpy.lib import recfunctions as rfn
from PIL import Image
import mediapipe as mp
from os import path
from scipy.spatial import Delaunay

from ..engine import Engine
from ...cleese import log
from ...third_party.mls.img_utils import mls_rigid_deformation


DLIB_TO_MEDIAPIPE = [
        162, 254, 132, 58, 172, 136, 150, 176,  # Face oval right
        152,  # Chin
        400, 379, 365, 397, 288, 361, 454, 389,  # Face oval left
        70, 63, 105, 66, 107,  # Eyebrow top right
        336, 296, 334, 293, 300,  # Eyebrow top left
        168, 187, 5, 1, 98, 97, 2, 326, 327,  # Nose
        33, 160, 158, 133, 153, 144,  # Eye right
        362, 385, 387, 263, 373, 380,  # Eye left
        61, 40, 37, 0, 267, 270, 291, 321, 314, 17, 84, 91,  # Outer lips
        78, 80, 13, 311, 308, 402, 14, 178,  # Inner lips
]

LANDMARKS_SETS = {
        "dlib-eyebrow-right": DLIB_TO_MEDIAPIPE[17:22],
        "dlib-eyebrow-left": DLIB_TO_MEDIAPIPE[22:27],
        "dlib-nose": DLIB_TO_MEDIAPIPE[27:36],
        "dlib-eye-right": DLIB_TO_MEDIAPIPE[36:42],
        "dlib-eye-left": DLIB_TO_MEDIAPIPE[42:48],
        "dlib-outer-lips": DLIB_TO_MEDIAPIPE[48:60],
        "dlib-inner-lips": DLIB_TO_MEDIAPIPE[60:68],
        "dlib-lips": DLIB_TO_MEDIAPIPE[48:68],
        "dlib-all": DLIB_TO_MEDIAPIPE[17:68],
        "mediapipe-all": list(np.arange(468))
}

MEDIAPIPE_TO_DLIB = [
        51,  # [0]
        30,  # [1]
        33,  # [2]
        -1, -1,
        29,  # [5]
        -1, -1, -1, -1, -1, -1, -1,
        62,  # [13]
        66,  # [14]
        -1, -1,
        57,  # [17]
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        36,  # [33]
        -1, -1, -1,
        50,  # [37]
        -1, -1,
        49,  # [40]
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        3,  # [58]
        -1, -1,
        48,  # [61]
        -1,
        18,  # [63]
        -1, -1,
        20,  # [66]
        -1, -1, -1,
        17,  # [70]
        -1, -1, -1, -1, -1, -1, -1,
        60,  # [78]
        -1,
        61,  # [80]
        -1, -1, -1,
        58,  # [84]
        -1, -1, -1, -1, -1, -1,
        59,  # [91]
        -1, -1, -1, -1, -1,
        32,  # [97]
        31,  # [98]
        -1, -1, -1, -1, -1, -1,
        19,  # [105]
        -1,
        21,  # [107]
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        2,  # [132]
        39,  # [133]
        -1, -1,
        5,  # [136]
        -1, -1, -1, -1, -1, -1, -1,
        41,  # [144]
        -1, -1, -1, -1, -1,
        6,  # [150]
        -1,
        8,  # [152]
        40,  # [153]
        -1, -1, -1, -1,
        38,  # [158]
        -1,
        37,  # [160]
        -1,
        0,  # [162]
        -1, -1, -1, -1, -1,
        27,  # [168]
        -1, -1, -1,
        4,  # [172]
        -1, -1, -1,
        7,  # [176]
        -1,
        67,  # [178]
        -1, -1, -1, -1, -1, -1, -1, -1,
        28,  # [187]
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        1,  # [254]
        -1, -1, -1, -1, -1, -1, -1, -1,
        45,  # [263]
        -1, -1, -1,
        52,  # [267]
        -1, -1,
        53,  # [270]
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        13,  # [288]
        -1, -1,
        54,  # [291]
        -1,
        25,  # [293]
        -1, -1,
        23,  # [296]
        -1, -1, -1,
        26,  # [300]
        -1, -1, -1, -1, -1, -1, -1,
        64,  # [308]
        -1, -1,
        63,  # [311]
        -1, -1,
        56,  # [314]
        -1, -1, -1, -1, -1, -1,
        55,  # [321]
        -1, -1, -1, -1,
        34,  # [326]
        35,  # [327]
        -1, -1, -1, -1, -1, -1,
        24,  # [334]
        -1,
        22,  # [336]
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        -1, -1, -1, -1, -1, -1,
        14,  # [361]
        42,  # [362]
        -1, -1,
        11,  # [365]
        -1, -1, -1, -1, -1, -1, -1,
        46,  # [373]
        -1, -1, -1, -1, -1,
        10,  # [379]
        47,  # [380]
        -1, -1, -1, -1,
        43,  # [385]
        -1,
        44,  # [387]
        -1,
        16,  # [389]
        -1, -1, -1, -1, -1, -1, -1,
        12,  # [397]
        -1, -1,
        9,  # [400]
        -1,
        65,  # [402]
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        15,  # [454]
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
]


def dlib_to_mediapipe(dlib_landmarks):
    return np.take(DLIB_TO_MEDIAPIPE, dlib_landmarks, axis=0)


def mediapipe_to_dlib(mediapipe_landmarks):
    dlib = np.take(MEDIAPIPE_TO_DLIB, mediapipe_landmarks, axis=0)
    if -1 in dlib:
        raise IndexError()
    return dlib


class FaceWarp(Engine):

    @staticmethod
    def load_file(filename):
        return np.array(Image.open(filename).convert("RGB")), {}

    @staticmethod
    def load_dfm(filename):
        dfm = None
        try:
            dfm = np.genfromtxt(filename,
                                dtype=[
                                    ("group", "i4"),
                                    ("index", "i4"),
                                    ("v1", "i4"),
                                    ("v2", "i4"),
                                    ("v3", "i4"),
                                    ("alpha", "f8"),
                                    ("beta", "f8"),
                                    ("gamma", "f8"),
                                ],
                                delimiter=",",
                                invalid_raise=True)
        except ValueError:
            log("ERROR: invalid dfm file: {}".format(filename))
        except OSError as e:
            log("ERROR: {}".format(e))
        return dfm

    @staticmethod
    def load_dfmxy(filename):
        dfm = None
        try:
            dfm = np.genfromtxt(filename,
                                dtype=[
                                    ("index", "i4"),
                                    ("posX", "i4"),
                                    ("posY", "i4"),
                                    ("defX", "f8"),
                                    ("defY", "f8"),
                                ],
                                delimiter=",",
                                invalid_raise=True)
        except ValueError:
            log("ERROR: invalid dfm file: {}".format(filename))
        except OSError as e:
            log("ERROR: {}".format(e))
        return dfm

    @staticmethod
    def load_landmarks(filename):
        lm = None
        try:
            lm = np.loadtxt(filename)
            if lm.ndim != 2 or lm.shape[1] != 2:
                log(("ERROR: {}, expected landmarks file with shape (N, 2), "
                     "got {}").format(filename, lm.shape))
                return None
        except OSError as e:
            log("ERROR: {}".format(e))
        return lm

    @staticmethod
    def process(img, config, file_output=False, dfm=None, dfmxy=None):

        # Check all the needed user-provided config values are here
        try:
            MLS_ALPHA = config["mediapipe"]["mls"]["alpha"]
            COVARIANCE_MAT = config["mediapipe"]["random_gen"]["covMat"]
            LANDMARKS_TYPES = config["mediapipe"]["random_gen"]["landmarks"]
            # NUM_FILES = config["main"]["numFiles"]
            FACE_THRESH = config["mediapipe"]["face_detect"]["threshold"]
        except KeyError as e:
            log("ERROR: missing config element: {}".format(e))
            return

        if dfm is not None and dfmxy is not None:
            log("WARN: non empty dfm and dfmxy provided. Ignoring dfmxy")

        mls_targets = []
        if dfm is not None:
            mls_targets += dlib_to_mediapipe(dfm["index"]).tolist()
            triangles_indices = dlib_to_mediapipe(
                    rfn.structured_to_unstructured(
                        dfm[["v1", "v2", "v3"]]))
        elif dfmxy is not None:
            mls_targets += dfmxy["index"].tolist()
        else:
            for key in LANDMARKS_TYPES:
                if key == "mediapipe":
                    mls_targets += LANDMARKS_TYPES[key]
                elif key == "dlib":
                    try:
                        mls_targets += dlib_to_mediapipe(
                                LANDMARKS_TYPES[key]).tolist()
                    except IndexError:
                        log("WARN: dlib landmark index outside of [0, 67]")
                elif key == "presets":
                    for preset in LANDMARKS_TYPES[key]:
                        if preset in LANDMARKS_SETS:
                            mls_targets += LANDMARKS_SETS[preset]
                        else:
                            log("WARN: unknown landmark preset: {}".format(
                                    preset))
                else:
                    log("ERROR: unknown landmark indexing: {}".format(key))

        # Retrieve face landmarks from Mediapipe
        results = None
        with mp.solutions.face_mesh.FaceMesh(
                static_image_mode=True,
                max_num_faces=1,
                refine_landmarks=True,
                min_detection_confidence=FACE_THRESH) as face_mesh:

            results = face_mesh.process(img)

        # Get points for the MLS
        if not results or not results.multi_face_landmarks:
            log("ERROR: unable to detect any faces")
            return

        rng = np.random.default_rng()

        # Retrieve targeted landmarks for Mediapipe's results
        height, width, _ = img.shape
        for face_landmarks in results.multi_face_landmarks:
            # Compute pixel size of the detected face
            x = [int(lm.x * width) for lm in face_landmarks.landmark]
            y = [int(lm.y * height) for lm in face_landmarks.landmark]
            np_landmarks = np.column_stack((x, y))
            face_min = np.amin(np_landmarks, axis=0)
            face_max = np.amax(np_landmarks, axis=0)
            face_size = face_max[1] - face_min[1]

            # Select landmarks to be deformed
            p = np.take(np_landmarks, np.array(mls_targets), axis=0)

            if dfm is not None:
                # Apply barycentric coordinates to get target points
                t = np.take(np_landmarks, triangles_indices, axis=0)
                weights = rfn.structured_to_unstructured(
                        dfm[["alpha", "beta", "gamma"]])
                q = np.sum(t * weights[:, :, np.newaxis], axis=1)
                q = q.astype(np.int16)
            elif dfmxy is not None:
                mls_offsets = rfn.structured_to_unstructured(
                        dfmxy[["defX", "defY"]])
                q = p + mls_offsets
            else:
                # Compute gaussian-distributed offsets
                mls_offsets = rng.multivariate_normal((0, 0),
                                                      COVARIANCE_MAT,
                                                      len(mls_targets))
                mls_offsets *= face_size
                q = p + mls_offsets

        # Setup fixed anchor point on the image's corners for the MLS
        box = np.array([
            (0, 0),
            (width - 1, 0),
            (width - 1, height - 1),
            (0, height - 1),
        ])
        p = np.concatenate((p, box))
        q = np.concatenate((q, box))

        gridX = np.arange(0, width, dtype=np.int16)
        gridY = np.arange(0, height, dtype=np.int16)
        vy, vx = np.meshgrid(gridX, gridY)

        # mls_rigid_deformation expects coordinates in the (y, x) order
        p_mls = p[:, [1, 0]]
        q_mls = q[:, [1, 0]]
        rigid = mls_rigid_deformation(vy, vx,
                                      p_mls, q_mls,
                                      alpha=MLS_ALPHA)

        # Apply the computed deformation to the image
        deformed_img = np.ones_like(img)
        deformed_img[vx, vy] = img[tuple(rigid)]

        return deformed_img

    @staticmethod
    def generate_stimuli(img, config, **kwargs):

        # default configuration values
        try:
            DFMXY_EXT = config["main"]["param_ext"]
        except KeyError as e:
            DFMXY_EXT = '.txt' # default dfm extension is .txt

        # Check all the needed user-provided config values are here
        try:
            MLS_ALPHA = config["mediapipe"]["mls"]["alpha"]
            COVARIANCE_MAT = config["mediapipe"]["random_gen"]["covMat"]
            LANDMARKS_TYPES = config["mediapipe"]["random_gen"]["landmarks"]
            NUM_FILES = config["main"]["numFiles"]
            FACE_THRESH = config["mediapipe"]["face_detect"]["threshold"]
        except KeyError as e:
            log("ERROR: missing config element: {}".format(e))
            return

        # Internal values, should never be missing
        BASE_DIR = config["main"]["expBaseDir"]
        FILENAME = config["main"]["filename"]

        # Load all the different sources of landmark targets
        mls_targets = []
        for key in LANDMARKS_TYPES:
            if key == "mediapipe":
                mls_targets += LANDMARKS_TYPES[key]
            elif key == "dlib":
                try:
                    mls_targets += dlib_to_mediapipe(
                            LANDMARKS_TYPES[key]).tolist()
                except IndexError:
                    log("WARN: dlib landmark index outside of [0, 67]")
            elif key == "presets":
                for preset in LANDMARKS_TYPES[key]:
                    if preset in LANDMARKS_SETS:
                        mls_targets += LANDMARKS_SETS[preset]
                    else:
                        log("WARN: unknown landmark preset: {}".format(preset))
            else:
                log("ERROR: unknown landmark indexing: {}".format(key))

        rng = np.random.default_rng()

        # Retrieve face landmarks from Mediapipe
        results = None
        with mp.solutions.face_mesh.FaceMesh(
                static_image_mode=True,
                max_num_faces=1,
                refine_landmarks=True,
                min_detection_confidence=FACE_THRESH) as face_mesh:

            results = face_mesh.process(img)

        # Get points for the MLS
        if not results or not results.multi_face_landmarks:
            log("ERROR: unable to detect any faces")
            return

        # Retrieve targeted landmarks for Mediapipe's results
        height, width, _ = img.shape
        np_landmarks = []
        p = []
        for face_landmarks in results.multi_face_landmarks:
            # Compute pixel size of the detected face
            x = [int(lm.x * width) for lm in face_landmarks.landmark]
            y = [int(lm.y * height) for lm in face_landmarks.landmark]
            np_landmarks = np.column_stack((x, y))
            face_min = np.amin(np_landmarks, axis=0)
            face_max = np.amax(np_landmarks, axis=0)
            face_size = face_max[1] - face_min[1]

            # Select landmarks to be deformed
            for i, lm_idx in enumerate(mls_targets):
                pt = np_landmarks[lm_idx]
                p.append(pt)

        # Export landmarks positions to file
        basename = path.splitext(path.basename(FILENAME))
        output_landmarks_file = path.join(
                BASE_DIR, "{}.landmarks.txt".format(basename[0]))
        np.savetxt(output_landmarks_file,
                   np_landmarks,
                   fmt="%d",
                   header="landmarks X, Y")

        # Setup fixed anchor point on the image's corners for the MLS
        p.append([0, 0])
        p.append([width - 1, 0])
        p.append([width - 1, height - 1])
        p.append([0, height - 1])
        p = np.array(p)

        gridX = np.arange(0, width, dtype=np.int16)
        gridY = np.arange(0, height, dtype=np.int16)
        vy, vx = np.meshgrid(gridX, gridY)

        for i in range(NUM_FILES):
            log("Random face deformation [facewarp]: {}/{}"
                .format(i + 1, NUM_FILES))

            # Compute gaussian-distributed offsets
            mls_offsets = rng.multivariate_normal((0, 0),
                                                  COVARIANCE_MAT,
                                                  len(mls_targets))
            mls_offsets *= face_size
            q = p[:-4] + mls_offsets
            q = np.concatenate((q, p[-4:]))

            # mls_rigid_deformation expects coordinates in the (y, x) order
            p_mls = p[:, [1, 0]]
            q_mls = q[:, [1, 0]]
            rigid = mls_rigid_deformation(vy, vx,
                                          p_mls, q_mls,
                                          alpha=MLS_ALPHA)

            # Apply the computed deformation to the image
            deformed_img = np.ones_like(img)
            deformed_img[vx, vy] = img[tuple(rigid)]

            # Export deformed image to file
            output_img_file = path.join(BASE_DIR, "{}.{:08d}{}".format(
                    basename[0], i + 1, basename[1]))
            out_image = Image.fromarray(deformed_img)
            out_image.save(output_img_file)

            # Export deformed landmarks to file
            output_dfm_file = path.join(BASE_DIR, "{}.{:08d}{}".format(
                    basename[0], i + 1, DFMXY_EXT))
            dfmxy = np.hstack((np.array([mls_targets]).T, p[:-4], mls_offsets))
            np.savetxt(output_dfm_file,
                       dfmxy,
                       fmt="%d,%d,%d,%.8f,%.8f",
                       header="idx, posX, posY, defX, defY", 
                       comments="")

    @staticmethod
    def dfmxy_to_dfm(dfmxy_file, landmarks_file, output_dfm_file=None):
        dfmxy = FaceWarp.load_dfmxy(dfmxy_file)
        if dfmxy is None:
            return

        landmarks = FaceWarp.load_landmarks(landmarks_file)
        if landmarks is None:
            return

        if output_dfm_file is None:
            output_dfm_file = path.splitext(dfmxy_file)[0] + ".dfm"

        dfmxy_pts = rfn.structured_to_unstructured(dfmxy[["posX", "posY"]])
        dfmxy_def = rfn.structured_to_unstructured(dfmxy[["defX", "defY"]])
        targets = dfmxy_pts + dfmxy_def

        try:
            dlib_idx = mediapipe_to_dlib(dfmxy["index"])
        except IndexError:
            log(("ERROR: dfmxy has indices without dlib counterparts. "
                 "Cannot convert."))
            return

        dlib_landmarks = np.take(landmarks, DLIB_TO_MEDIAPIPE, axis=0)
        tri = Delaunay(dlib_landmarks)
        target_triangles = tri.find_simplex(targets, bruteforce=True, tol=0.01)

        dfm = np.empty((len(dlib_idx), 8))
        trans_targets = targets - tri.transform[target_triangles, 2]
        for i, t in enumerate(target_triangles):
            b = tri.transform[t, :2].dot(np.transpose(trans_targets[i]))
            # group all landmarks in group 0
            # better than trying to guess at group composition
            dfm[i] = np.concatenate(([0, dlib_idx[i]],
                                     tri.simplices[t],
                                     np.transpose(b),
                                     [1 - b.sum(axis=0)]))

        np.savetxt(output_dfm_file,
                   dfm,
                   fmt="%d,%d,%d,%d,%d,%.8f,%.8f,%.8f",
                   header=("group, index, triangle indices {0, 1, 2}, "
                           "barycentric coords weights {alpha, beta, gamma}"))

    @staticmethod
    def name():
        return "facewarp"

    @staticmethod
    def img_read(file_name): 
        return FaceWarp.load_file(file_name)[0]

    @staticmethod
    def img_write(img_array, file_name): 
        img = Image.fromarray(img_array)
        img.save(file_name)





#!/usr/bin/env python
'''
CLEESE toolbox v1.0
mar 2018, J.J. Burred <jjburred@jjburred.com> for IRCAM/CNRS

Spectral processing engine: STFT with time-varying hops, phase-locking phase vocoder and ISTFT with window-by-window resampling.
'''

import numpy as np
import scipy.fftpack as fft
import scipy
import scipy.signal as sig
import scipy.interpolate


def stft(waveForm,win,winSize,hopVec):

    numSamples = waveForm.shape[0]
    numFrames = len(hopVec)

    # security padding
    numExtraSamples = np.sum(hopVec) - numSamples + winSize*2
    waveForm = np.concatenate((waveForm,np.zeros(numExtraSamples)))

    atoms = np.zeros((winSize,numFrames))

    pos = 0
    for i in range(0,numFrames):
        atoms[:,i] = waveForm[pos:pos+winSize]*win
        pos += hopVec[i]

    stft = np.fft.fft(atoms,axis=0)

    return stft

def istft(stft,win,winSize,hopSize):

    numFrames = stft.shape[1]
    numSamples = numFrames*hopSize + winSize

    outWaveform = np.zeros(numSamples) # TODO: stereo

    winSum = np.zeros(numSamples)
    winSq = win * win

    pos = 0
    for i in range(0,numFrames):

        currSpec = stft[:, i].flatten()
        currSpec = np.concatenate((currSpec, currSpec[-2:0:-1].conj()), 0)

        outWaveform[pos:(pos+winSize)] += fft.ifft(currSpec).real * win
        winSum[pos:(pos+winSize)]      += winSq

        pos += hopSize

    # amplitude normalization
    nonZero = winSum > np.finfo(winSum.dtype).tiny
    outWaveform[nonZero] /= winSum[nonZero]

    return outWaveform

def istft_resamp(stft,win,winSize,hopSize,ratesVec,oriLength):

    numFrames = stft.shape[1]
    numSamples = oriLength + 4*winSize

    outWaveform = np.zeros(numSamples)

    winSum = np.zeros(numSamples)
    winSq = win * win

    pos = 0
    for i in range(0,numFrames):

        currSpec = stft[:, i].flatten()
        currSpec = np.concatenate((currSpec, currSpec[-2:0:-1].conj()), 0)
        currAtom = fft.ifft(currSpec).real * win

        n_currLen = int(round(winSize/ratesVec[i]))

        if (pos+n_currLen) > numSamples:
            break

        outWaveform[pos:(pos+n_currLen)] += sig.resample(currAtom,n_currLen)
        winSum[pos:(pos+n_currLen)]      += sig.resample(winSq,n_currLen)

        pos += int(round(hopSize/ratesVec[i]))

    # amplitude normalization
    nonZero = winSum > np.finfo(winSum.dtype).tiny
    outWaveform[nonZero] /= winSum[nonZero]

    return outWaveform


def phaseVocoder_varHop(specIn, anaHopVec, synHop=256, phase_locking=True):

    specOut = np.zeros(specIn.shape,dtype=complex)
    specOut[:,0] = specIn[:,0]

    n_frames = specIn.shape[1]
    halfLen = specIn.shape[0]
    fftLen = (halfLen-1)*2
    omega = 2 * np.pi * np.arange(0,halfLen) / fftLen

    for i in range(1,n_frames):

        currPhase = np.angle(specIn[:,i])
        lastPhase = np.angle(specIn[:,i-1])

        expAdv = omega * anaHopVec[i-1]  # expected phase advance

        deltaPhase = currPhase - lastPhase - expAdv
        deltaPhase = deltaPhase - 2 * np.pi * np.round(deltaPhase/(2*np.pi))  # princarg function

        hopAdv = (omega + deltaPhase/float(anaHopVec[i])) * synHop  # phase advance during synth hop

        lastOutPhase = np.angle(specOut[:,i-1])

        if phase_locking:

            peakIndVec, regStart, regEnd = findPeaks(np.abs(specIn[:,i]))

            theta = np.zeros(len(specOut[:,i]))
            for pk, st, en in zip(peakIndVec, regStart, regEnd):
                theta[st:en+1] = lastOutPhase[pk] + hopAdv[pk] - currPhase[pk]

        else:

            theta = lastOutPhase + hopAdv - currPhase


        specOut[:,i] = np.exp(1j * theta) * specIn[:,i]


    return specOut


def findPeaks(inVec):

    nn = 4 # number of neighbors. must be even

    flagMat = np.zeros((nn,len(inVec)))

    # zero pad
    oriVec = inVec
    inVec = np.append(inVec,np.zeros(nn//2))
    inVec = np.insert(inVec,0,np.zeros(nn//2))

    # find peaks
    startInd = np.arange(0,nn+1)
    startInd = np.delete(startInd,nn//2)
    endInd = np.arange(-nn,1)
    endInd = np.delete(endInd,nn//2)
    endInd[-1] = len(inVec)
    for i,(st,en) in enumerate(zip(startInd,endInd)):
        flagMat[i,:] = inVec[st:en] < inVec[nn//2:-nn//2]
    peakIndVec = np.nonzero(np.prod(flagMat,axis=0))[0]

    if len(peakIndVec) == 0:
        return [],[],[]

    # define influence regions
    regStart = np.zeros(len(peakIndVec),dtype=int)
    regEnd   = np.zeros(len(peakIndVec),dtype=int)

    regStart[0] = 0
    regStart[1:] = np.ceil((peakIndVec[1:] + peakIndVec[0:-1])/2)
    regEnd[0:-1] = regStart[1:] - 1
    regEnd[-1] = len(oriVec)


    return peakIndVec, regStart, regEnd



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
CLEESE toolbox
v1.0: mar 2018, J.J. Burred <jjburred@jjburred.com> for IRCAM/CNRS
v2.0: jan 2022, Lara Kermarec <lara.git@kermarec.bzh> for CNRS

Audio utils functions for CLEESE's phase vocoder engine
'''

import scipy
import numpy as np
import math

from ...third_party.yin import compute_yin


def load_file(file_name):

    sampleRate, wave = scipy.io.wavfile.read(file_name)
    sampleFormat = wave.dtype
    if sampleFormat in ('int16', 'int32'):
        # convert to float
        if sampleFormat == 'int16':
            n_bits = 16
        elif sampleFormat == 'int32':
            n_bits = 32
        wave = wave/(float(2**(n_bits - 1)))
        wave = wave.astype('float32')

    attributes = {
        "sample_rate": sampleRate,
        "sample_format": sampleFormat,
    }
    return wave, attributes

def wav_read(file_name):
    wave, attr = load_file(file_name)
    return wave, attr["sample_rate"], attr["sample_format"]

def wav_write(wave_out, file_name, sr, sample_format='int16'):

    if sample_format == 'int16':
        wave_out_format = wave_out * 2**15
    elif sample_format == 'int32':
        wave_out_format = wave_out * 2**31
    else:
        wave_out_format = wave_out
    wave_out_format = wave_out_format.astype(sample_format)
    scipy.io.wavfile.write(file_name, sr, wave_out_format)

def extract_pitch(x, sr, win=.02, bounds=[70,400], interpolate=True):
    """
    Extract pitch from x waveform with the YIN algorithm
    """
    
    # extract raw pitch with yin
    hop_size = int(np.floor(sr * win))
    min_f0, max_f0 = bounds
    # YIN assumes the analysis window is longer than the min_f0 period 
    if hop_size < int(sr / min_f0): 
        hop_size = int(sr / min_f0) + 1
    pitch, harmonic_rates, argmins, times = compute_yin(x, sr, 
        dataFileName=None, w_len=hop_size, w_step=hop_size, f0_min=min_f0, f0_max=max_f0, harmo_thresh=0.1)
    
    # third-party yin returns lists; convert to nparray
    pitch = np.array(pitch)
    times = np.array(times)

    # clean up unvoiced areas
    # flag unvoiced as nan
    pitch[np.where(pitch == 0)[0]] = np.nan
    # trim beginning and end nans
    notnans = np.flatnonzero(~np.isnan(pitch))
    if notnans.size:
        pitch = pitch[notnans[0]: notnans[-1]+1]
        times = times[notnans[0]: notnans[-1]+1]
    else: 
        pitch = times = []
    
    if(interpolate):  # interpolate nans
        start_value = end_value = np.mean(pitch[np.nonzero(~np.isnan(pitch))])
        pitch = interpolate_pitch(pitch, start_value = start_value, end_value = end_value)
    
    return times, pitch

def interpolate_pitch(x, start_value, end_value): 
    """
    Interpolate zeros in pitch series. Method = spline (order-3 polynomial), linear or none. 
    Provide start_value, end_value to fix bounds. 
    """
    xp=np.where(np.invert(list(map(math.isnan, x))))[0]
    fp=np.array(x)[np.where(np.invert(list(map(math.isnan, x))))]
    return np.interp(x=range(len(x)), xp=xp,fp=fp)

def lpc_env(x, sr, order=6):
    """
    Compute spectral enveloppe with linear prediction coefficient (LPC) algorithm. 
    order is the number of poles (rule of thumb: 6 poles to extract 3 formants)
    """
    
    n_samples=len(x)

    #compute Mth-order autocorrelation function
    rx=[]
    for i in range(order+1):
        rx.append(x[:n_samples-i].dot(x[i:n_samples]))

    #construct Toeplitz matrix
    rx = np.array(rx)
    
    #solve toeplitz matrix
    a_coeffs = scipy.linalg.solve_toeplitz((-rx[0:order],-rx[0:order]),rx[1:order+1])

    #get complete polynomial A(z)
    a_lp = np.insert(a_coeffs,0,1)
    
    #get curve
    freqs, env = scipy.signal.freqz(1, a_lp, fs=sr)
    env_db = 20*np.log10(abs(env))
    
    return freqs, env_db

def extract_spectral_env(x, sr, lpc_order=50, pe_thresh=1000, freq_limit = 20000):
    """
    Extract spectral envelope from waveform x using LPC algorithm
    Plot as plot(freqs,env_db)
    """
    # pre-emphasis filter s[n] = s[n] - a*s[n-1]
    a = np.exp(-2*np.pi*pe_thresh/sr)
    x_pe = scipy.signal.lfilter([1, -a], [1], x)
    
    # resample at 2*freq_limit
    new_sr = 2*freq_limit 
    new_n = round(len(x_pe) * new_sr / sr)
    x_rs = scipy.signal.resample(x_pe, new_n)
    
    # normalize
    x_rs = x_rs/max(abs(x_rs))
        
    # extract LPC envelope
    freqs, env_db = lpc_env(x_rs,new_sr,lpc_order)
    
    return freqs, env_db


#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
CLEESE toolbox v2.0
jan 2022, Lara Kermarec <lara.git@kermarec.bzh> for CNRS

CLEESE's engines API interface
'''

from abc import ABC, abstractmethod


class Engine(ABC):

    @staticmethod
    @abstractmethod
    def load_file(file_name):
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def process(data, config, **kwargs):
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def generate_stimuli(data, config, **kwargs):
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def name():
        raise NotImplementedError()



# Source: https://github.com/aliutkus/torchinterp1d
#
# Modification: Remove backward function
#

import torch


def interp(xnew, x, y):
    """
    Linear 1D interpolation on the GPU for Pytorch.
    This function returns interpolated values of a set of 1-D functions at
    the desired query points `xnew`.
    This function is working similarly to Matlab™ or scipy functions with
    the `linear` interpolation mode on, except that it parallelises over
    any number of desired interpolation problems.
    The code will run on GPU if all the tensors provided are on a cuda
    device.
    Parameters
    ----------
    x : (N, ) or (D, N) Pytorch Tensor
        A 1-D or 2-D tensor of real values.
    y : (N,) or (D, N) Pytorch Tensor
        A 1-D or 2-D tensor of real values. The length of `y` along its
        last dimension must be the same as that of `x`
    xnew : (P,) or (D, P) Pytorch Tensor
        A 1-D or 2-D tensor of real values. `xnew` can only be 1-D if
        _both_ `x` and `y` are 1-D. Otherwise, its length along the first
        dimension must be the same as that of whichever `x` and `y` is 2-D.
    """
    # making the vectors at least 2D
    is_flat = {}
    require_grad = {}
    v = {}
    device = []
    eps = torch.finfo(y.dtype).eps
    for name, vec in {'x': x, 'y': y, 'xnew': xnew}.items():
        assert len(vec.shape) <= 2, 'interp1d: all inputs must be '\
                                    'at most 2-D.'
        if len(vec.shape) == 1:
            v[name] = vec[None, :]
        else:
            v[name] = vec
        is_flat[name] = v[name].shape[0] == 1
        require_grad[name] = vec.requires_grad
        device = list(set(device + [str(vec.device)]))
    assert len(device) == 1, 'All parameters must be on the same device.'
    device = device[0]

    # Checking for the dimensions
    assert (v['x'].shape[1] == v['y'].shape[1]
            and (
                    v['x'].shape[0] == v['y'].shape[0]
                    or v['x'].shape[0] == 1
                    or v['y'].shape[0] == 1
                )
            ), ("x and y must have the same number of columns, and either "
                "the same number of row or one of them having only one "
                "row.")

    reshaped_xnew = False
    if ((v['x'].shape[0] == 1) and (v['y'].shape[0] == 1)
        and (v['xnew'].shape[0] > 1)):
        # if there is only one row for both x and y, there is no need to
        # loop over the rows of xnew because they will all have to face the
        # same interpolation problem. We should just stack them together to
        # call interp1d and put them back in place afterwards.
        original_xnew_shape = v['xnew'].shape
        v['xnew'] = v['xnew'].contiguous().view(1, -1)
        reshaped_xnew = True

    # identify the dimensions of output and check if the one provided is ok
    D = max(v['x'].shape[0], v['xnew'].shape[0])
    shape_ynew = (D, v['xnew'].shape[-1])

    ynew = torch.zeros(*shape_ynew, device=device)

    # moving everything to the desired device in case it was not there
    # already (not handling the case things do not fit entirely, user will
    # do it if required.)
    for name in v:
        v[name] = v[name].to(device)

    # calling searchsorted on the x values.
    ind = ynew.long()

    # expanding xnew to match the number of rows of x in case only one xnew is
    # provided
    if v['xnew'].shape[0] == 1:
        v['xnew'] = v['xnew'].expand(v['x'].shape[0], -1)

    torch.searchsorted(v['x'].contiguous(),
                        v['xnew'].contiguous(), out=ind)

    # the `-1` is because searchsorted looks for the index where the values
    # must be inserted to preserve order. And we want the index of the
    # preceeding value.
    ind -= 1
    # we clamp the index, because the number of intervals is x.shape-1,
    # and the left neighbour should hence be at most number of intervals
    # -1, i.e. number of columns in x -2
    ind = torch.clamp(ind, 0, v['x'].shape[1] - 1 - 1)

    # helper function to select stuff according to the found indices.
    def sel(name):
        if is_flat[name]:
            return v[name].contiguous().view(-1)[ind]
        return torch.gather(v[name], 1, ind)

    # activating gradient storing for everything now
    enable_grad = False
    saved_inputs = []
    for name in ['x', 'y', 'xnew']:
        if require_grad[name]:
            enable_grad = True
            saved_inputs += [v[name]]
        else:
            saved_inputs += [None, ]
    # assuming x are sorted in the dimension 1, computing the slopes for
    # the segments
    is_flat['slopes'] = is_flat['x']
    # now we have found the indices of the neighbors, we start building the
    # output. Hence, we start also activating gradient tracking
    v['slopes'] = (
            (v['y'][:, 1:]-v['y'][:, :-1])
            /
            (eps + (v['x'][:, 1:]-v['x'][:, :-1]))
        )

    # now build the linear interpolation
    ynew = sel('y') + sel('slopes') * (v['xnew'] - sel('x'))

    if reshaped_xnew:
        ynew = ynew.view(original_xnew_shape)

    return ynew



#!/usr/bin/env python
'''
CLEESE toolbox v1.0
mar 2018, J.J. Burred <jjburred@jjburred.com> for IRCAM/CNRS

Breakpoint function (BPF) creation and utility functions.
'''

import numpy as np


def createBPF(tr, config, BPFtime, numPoints, endOnTrans, eqFreqVec=None):

    # for pitch, stretch and gain, BPFval is a vector
    if tr in ["pitch", "stretch", "gain"]:

        BPFval = createBPFamplVec(config[tr], numPoints, endOnTrans)

        if tr == "stretch":
            BPFval += 1
            BPFval[BPFval <= 0] = 1
        elif tr == "gain":
            BPFval = 10**(BPFval/10.0)

        BPF = np.column_stack((BPFtime, BPFval))

    # for EQ, BPFval is a matrix
    elif tr == "eq":

        numBands = config["eq"]["band"]["count"] + 1

        BPFval = np.zeros((numBands, len(BPFtime)))
        for i in range(0, numBands):
            BPFval[i, :] = createBPFamplVec(config["eq"], numPoints, 0)

        # line format of an EQ BPF: time numBands freq1 ampl1 freq2 ampl2 ...
        BPF = np.zeros((len(BPFtime), (config["eq"]["band"]["count"]+1)*2 + 2))
        for i in range(0, len(BPFtime)):
            BPF[i, 0] = BPFtime[i]
            BPF[i, 1] = config["eq"]["band"]["count"]
            BPF[i, 2::2] = eqFreqVec
            BPF[i, 3::2] = BPFval[:, i]

    return BPF

def create_BPF_header(tr, config): 
    ''' 
        returns a str that's the header for the file destined to store BPF values
    '''

    # for pitch, stretch and gain, BPF file is two-columned in the form time,value
    if tr in ["pitch", "stretch", "gain"]:
        header = "time,%s"%tr
    # for eq, BPF file is in the form: time numBands freq1 ampl1 freq2 ampl2 ...
    elif tr == "eq":
        num_bands = config["eq"]["band"]["count"] + 1 # one more cutoff frequency than bands
        header = "time,n_channels"
        for band in range(0, num_bands):
            header += ",freq_%d,ampl_%d"%(band,band)
    return header


def createBPFtimeVec(duration, local_pars, timeVec=None):

    BPFtime = np.array([0])
    numPoints = 1
    endOnTrans = 0

    if timeVec is not None:
        numWin = len(timeVec)
    elif local_pars["window"]["unit"] == 'n':
        if local_pars["window"]["count"] == 0:
            return BPFtime, numPoints, endOnTrans
        numWin = local_pars["window"]["count"]
    elif local_pars["window"]["unit"] == 's':
        if local_pars["window"]["len"] == 0:
            return BPFtime, numPoints, endOnTrans
        numWin = np.floor(duration/local_pars["window"]["len"]) + 1

    if numWin == 1:
        return BPFtime, numPoints, endOnTrans

    if timeVec is not None:
        BPFtime = timeVec
        if timeVec[0] != 0:
            BPFtime = np.insert(BPFtime, 0, 0.)
            numWin += 1
        BPFtime = np.append(BPFtime, duration)
    elif local_pars["window"]["unit"] == 'n':
        BPFtime = np.linspace(0, duration, numWin+1)
    elif local_pars["window"]["unit"] == 's':
        BPFtime = local_pars["window"]["len"]*np.arange(0, numWin)
        BPFtime = np.append(BPFtime, duration)

    if local_pars['BPFtype'] == 'ramp':
        numPoints = numWin + 1
    elif local_pars['BPFtype'] == 'square':
        BPFtime = np.sort(np.concatenate((
            BPFtime[1:-1]-local_pars['trTime']/2,
            BPFtime[1:-1]+local_pars['trTime']/2)))
        if BPFtime[-1] > duration:
            BPFtime = np.delete(temp, -1)
            endOnTrans = 1
        BPFtime = np.append(BPFtime, duration)
        BPFtime = np.insert(BPFtime, 0, 0.)
        numPoints = numWin

    return BPFtime, numPoints, endOnTrans


def createBPFamplVec(local_pars, numPoints, endOnTrans):

    # random BPF values
    BPFval = np.random.normal(0, local_pars['std'], int(numPoints))

    # eliminate far samples (trunc factor)
    # and replace them by new random values
    for i in range(0, len(BPFval)):
        while np.abs(BPFval[i]) > (local_pars['std'] * local_pars['trunc']):
            BPFval[i] = np.random.normal(0, local_pars['std'], 1)

    if numPoints == 1:
        return BPFval

    if local_pars['BPFtype'] == 'square':
        temp = np.array([])
        for i in range(0, len(BPFval)):
            temp = np.append(temp, [BPFval[i], BPFval[i]])
        if endOnTrans:
            temp = np.delete(temp, -1)
        BPFval = temp

    return(BPFval)


def createBPFfreqs(config):

    sr = config["main"]["sr"]
    band_count = config["eq"]["band"]["count"]

    if config["eq"]["scale"] == "linear":
        min_freq = float(config["eq"]["min_freq"])
        max_freq = float(config["eq"]["max_freq"])
        eqFreqVec = np.linspace(min_freq, max_freq, band_count+1)

    elif config["eq"]["scale"] == "mel":
        minMel = freq2mel(config["eq"]["min_freq"])
        if np.isnan(config["eq"]["max_freq"]):
            maxMel = freq2mel(sr/float(2))
        else:
            maxMel = freq2mel(config["eq"]["max_freq"])

        melPerBin = (maxMel-minMel)/(band_count)
        eqFreqVec = melPerBin*np.ones(band_count)
        eqFreqVec = np.insert(eqFreqVec, 0, minMel)
        Vmel2freq = np.vectorize(mel2freq)
        eqFreqVec = Vmel2freq(np.cumsum(eqFreqVec))

    return eqFreqVec


def freq2mel(freq):
    return 1127.01048*np.log(1.+freq/700.)


def mel2freq(mel):
    return 700.0*(np.exp(mel/1127.01048)-1.0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
CLEESE toolbox v2.0
jan 2022, Lara Kermarec <lara.git@kermarec.bzh> for CNRS

External API for CLEESE
'''

import tomli
import os
import shutil
import time
import sys


def log(*args):
    print(*args, file=sys.stderr)


def process_file(engine, filename, config_file, **kwargs):
    data, attr = engine.load_file(filename)
    return process_data(engine, data, config_file, **attr, **kwargs)


def process_data(engine, data, config_file, **kwargs):
    conf = load_config(config_file)
    return engine.process(data, conf, **kwargs)


def generate_stimuli(engine, filename, config_file, **kwargs):
    conf = load_config(config_file)

    # Check all the needed user-provided config values are here
    try:
        OUT_PATH = conf["main"]["outPath"]
    except KeyError as e:
        log("ERROR: missing config element: {}".format(e))
        return

    # generate experiment name and folder
    main = conf["main"]
    if "generateExpFolder" in main and main["generateExpFolder"]:
        base_dir = os.path.join(OUT_PATH, "[{}]_{}".format(
                engine.name(), time.strftime("%Y-%m-%d_%H-%M-%S")))
    else:
        base_dir = OUT_PATH

    if not os.path.exists(base_dir):
        os.makedirs(base_dir)

    conf["main"]["expBaseDir"] = base_dir
    conf["main"]["filename"] = filename

    # copy original data to experiment folder
    shutil.copy2(filename, base_dir)

    # copy configuration file to experiment folder
    shutil.copy2(config_file, base_dir)

    data, attr = engine.load_file(filename)
    return engine.generate_stimuli(data, conf, **attr, **kwargs)


def load_config(filename):
    conf = None
    with open(filename, "rb") as f:
        try:
            conf = tomli.load(f)
        except tomli.TOMLDecodeError as e:
            log("Could not load configuration from '{}':\n  {}"
                .format(filename, e))
    return conf


if __name__ == "__main__":
    config_file = "./cleese-vanilla.toml"
    load_config(config_file)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Image deformation using moving least squares

@author: Jian-Wei ZHANG
@email: zjw.cs@zju.edu.cn
@date: 2017/8/8
@update: 2020/9/25
@update: 2021/7/14: Simplify usage
@update: 2021/12/24: Fix bugs and add an example of random control points (see `demo2()`)
"""

import time

import numpy as np
import matplotlib.pyplot as plt

try:
    import torch    # Install PyTorch first: https://pytorch.org/get-started/locally/
    from img_utils_pytorch import (
        mls_affine_deformation as mls_affine_deformation_pt,
        mls_similarity_deformation as mls_similarity_deformation_pt,
        mls_rigid_deformation as mls_rigid_deformation_pt,
    )
    device = torch.device("cuda:0" if torch.cuda.is_available() else 'cpu')
except ImportError as e:
    print(e)

from img_utils import (
    mls_affine_deformation,
    mls_similarity_deformation,
    mls_rigid_deformation
)

from PIL import Image


def demo():
    p = np.array([
        [155, 30], [155, 125], [155, 225],
        [235, 100], [235, 160], [295, 85], [293, 180]
    ])
    q = np.array([
        [211, 42], [155, 125], [100, 235],
        [235, 80], [235, 140], [295, 85], [295, 180]
    ])

    image = np.array(Image.open("images/toy.jpg"))

    height, width, _ = image.shape
    gridX = np.arange(width, dtype=np.int16)
    gridY = np.arange(height, dtype=np.int16)
    vy, vx = np.meshgrid(gridX, gridY)

    affine = mls_affine_deformation(vy, vx, p, q, alpha=1)
    aug1 = np.ones_like(image)
    aug1[vx, vy] = image[tuple(affine)]

    similar = mls_similarity_deformation(vy, vx, p, q, alpha=1)
    aug2 = np.ones_like(image)
    aug2[vx, vy] = image[tuple(similar)]

    rigid = mls_rigid_deformation(vy, vx, p, q, alpha=1)
    aug3 = np.ones_like(image)
    aug3[vx, vy] = image[tuple(rigid)]

    fig, ax = plt.subplots(1, 4, figsize=(12, 4))
    ax[0].imshow(image)
    ax[0].set_title("Original Image")
    ax[1].imshow(aug1)
    ax[1].set_title("Affine Deformation")
    ax[2].imshow(aug2)
    ax[2].set_title("Similarity Deformation")
    ax[3].imshow(aug3)
    ax[3].set_title("Rigid Deformation")

    for x in ax.flat:
        x.axis("off")

    plt.tight_layout(w_pad=0.1)
    plt.show()


def demo_torch():
    p = torch.from_numpy(np.array([
        [155, 30], [155, 125], [155, 225],
        [235, 100], [235, 160], [295, 85], [293, 180]
    ])).to(device)
    q = torch.from_numpy(np.array([
        [211, 42], [155, 125], [100, 235],
        [235, 80], [235, 140], [295, 85], [295, 180]
    ])).to(device)

    image = torch.from_numpy(np.array(Image.open("images/toy.jpg"))).to(device)

    height, width, _ = image.shape
    gridX = torch.arange(width, dtype=torch.int16).to(device)
    gridY = torch.arange(height, dtype=torch.int16).to(device)
    vy, vx = torch.meshgrid(gridX, gridY)
    # !!! Pay attention !!!: the shape of returned tensors are different between numpy.meshgrid and torch.meshgrid
    vy, vx = vy.transpose(0, 1), vx.transpose(0, 1)

    affine = mls_affine_deformation_pt(vy, vx, p, q, alpha=1)
    aug1 = torch.ones_like(image).to(device)
    aug1[vx.long(), vy.long()] = image[tuple(affine)]

    similar = mls_similarity_deformation_pt(vy, vx, p, q, alpha=1)
    aug2 = torch.ones_like(image).to(device)
    aug2[vx.long(), vy.long()] = image[tuple(similar)]

    rigid = mls_rigid_deformation_pt(vy, vx, p, q, alpha=1)
    aug3 = torch.ones_like(image).to(device)
    aug3[vx.long(), vy.long()] = image[tuple(rigid)]

    fig, ax = plt.subplots(1, 4, figsize=(12, 4))
    ax[0].imshow(image)
    ax[0].set_title("Original Image")
    ax[1].imshow(aug1.cpu().numpy())
    ax[1].set_title("Affine Deformation")
    ax[2].imshow(aug2.cpu().numpy())
    ax[2].set_title("Similarity Deformation")
    ax[3].imshow(aug3.cpu().numpy())
    ax[3].set_title("Rigid Deformation")

    for x in ax.flat:
        x.axis("off")

    plt.tight_layout(w_pad=0.1)
    plt.show()


def demo2():
    """ Smiled Monalisa """
    np.random.seed(1234)

    image = np.array(Image.open("images/monalisa.jpg"))
    height, width, _ = image.shape

    # Define deformation grid
    gridX = np.arange(width, dtype=np.int16)
    gridY = np.arange(height, dtype=np.int16)
    vy, vx = np.meshgrid(gridX, gridY)

    # ================ Control points group 1 (manually specified) ==================
    p1 = np.array([[0, 0], [517, 0], [0, 798], [517, 798],
        [140, 186], [135, 295], [181, 208], [181, 261], [203, 184], [202, 304], [225, 213],
        [225, 243], [244, 211], [244, 253], [254, 195], [281, 232], [252, 285]
    ])
    q1 = np.array([[0, 0], [517, 0], [0, 798], [517, 798],
        [140, 186], [135, 295], [181, 208], [181, 261], [203, 184], [202, 304], [225, 213],
        [225, 243], [238, 207], [237, 261], [253, 199], [281, 232], [249, 279]
    ])

    rigid1 = mls_rigid_deformation(vy, vx, p1, q1, alpha=1)
    aug1 = np.ones_like(image)
    aug1[vx, vy] = image[tuple(rigid1)]

    # ====================== Control points group 1 (random) =======================
    p2 = np.stack((
        np.random.randint(0, height, size=13),
        np.random.randint(0, width, size=13),
    ), axis=1)
    q2 = p2 + np.random.randint(-20, 20, size=p2.shape)

    rigid2 = mls_rigid_deformation(vy, vx, p2, q2, alpha=1)
    aug2 = np.ones_like(image)
    aug2[vx, vy] = image[tuple(rigid2)]

    fig, ax = plt.subplots(1, 3, figsize=(13, 6))
    ax[0].imshow(image)
    ax[0].set_title("Original Image")
    ax[1].imshow(aug1)
    ax[1].set_title("Manually specified control points")
    ax[2].imshow(aug2)
    ax[2].set_title("Random control points")

    for x in ax.flat:
        x.axis("off")

    plt.tight_layout(w_pad=1.0, h_pad=1.0)
    plt.show()


def read_tif(frame):
    image_pil = Image.open("images/train-volume.tif")
    image_pil.seek(frame)
    image = np.array(image_pil)
    label_pil = Image.open("images/train-labels.tif")
    label_pil.seek(frame)
    label = np.array(label_pil)

    return image, label


def demo3():
    image, label = read_tif(1)
    image = np.pad(image, ((30, 30), (30, 30)), mode='symmetric')
    label = np.pad(label, ((30, 30), (30, 30)), mode='symmetric')

    height, width = image.shape
    gridX = np.arange(width, dtype=np.int16)
    gridY = np.arange(height, dtype=np.int16)
    vy, vx = np.meshgrid(gridX, gridY)

    def augment(p, q, mode='affine'):
        if mode.lower() == 'affine':
            transform = mls_affine_deformation(vy, vx, p, q, alpha=1)
        elif mode.lower() == 'similar':
            transform = mls_similarity_deformation(vy, vx, p, q, alpha=1)
        elif mode.lower() == 'rigid':
            transform = mls_rigid_deformation(vy, vx, p, q, alpha=1)
        else:
            raise ValueError

        aug_img = np.ones_like(image)
        aug_img[vx, vy] = image[tuple(transform)]
        aug_lab = np.ones_like(label)
        aug_lab[vx, vy] = label[tuple(transform)]

        return aug_img, aug_lab

    fig, ax = plt.subplots(2, 4, figsize=(12, 6))
    ax[0, 0].imshow(image, cmap='gray')
    ax[0, 0].set_title("Original Image")
    ax[1, 0].imshow(label, cmap='gray')
    ax[1, 0].set_title("Original Label")

    np.random.seed(1234)
    p = np.c_[np.random.randint(0, height, size=32), np.random.randint(0, width, size=32)]
    q = p + np.random.randint(-15, 15, size=p.shape)
    q[:, 0] = np.clip(q[:, 0], 0, height)
    q[:, 1] = np.clip(q[:, 1], 0, width)
    p = np.r_[p, np.array([[0, 0], [0, width - 1], [height - 1, 0], [height - 1, width - 1]])]  # fix corner points
    q = np.r_[q, np.array([[0, 0], [0, width - 1], [height - 1, 0], [height - 1, width - 1]])]  # fix corner points

    for i, mode in enumerate(['Affine', 'Similar', 'Rigid']):
        aug_img, aug_lab = augment(p, q, mode)
        ax[0, i + 1].imshow(aug_img, cmap='gray')
        ax[0, i + 1].set_title(f"{mode} Deformated Image")
        ax[1, i + 1].imshow(aug_lab, cmap='gray')
        ax[1, i + 1].set_title(f"{mode} Deformated Label")

    for x in ax.flat:
        x.axis('off')

    plt.tight_layout(w_pad=1.0, h_pad=1.0)
    plt.show()


def benchmark_numpy(image, p, q):
    height, width = image.shape[:2]

    # Define deformation grid
    gridX = np.arange(width, dtype=np.int16)
    gridY = np.arange(height, dtype=np.int16)
    vy, vx = np.meshgrid(gridX, gridY)

    rigid = mls_rigid_deformation(vy, vx, p, q, alpha=1)
    aug = np.ones_like(image)
    aug[vx, vy] = image[tuple(rigid)]
    return aug


def benchmark_torch(image, p, q):
    height, width = image.shape[:2]
    device = image.device

    # Define deformation grid
    gridX = torch.arange(width, dtype=torch.int16).to(device)
    gridY = torch.arange(height, dtype=torch.int16).to(device)
    vy, vx = torch.meshgrid(gridX, gridY)

    rigid = mls_rigid_deformation_pt(vy, vx, p, q, alpha=1)
    aug = torch.ones_like(image).to(device)
    aug[vx.long(), vy.long()] = image[rigid[0], rigid[1]]
    return aug


def run_benckmark(i):
    sizes = [   # (height, width)
        (100, 100),
        (500, 500),
        (500, 500),
        (500, 500),
        (1000, 1000),
        (2000, 2000),
    ]
    num_pts = [16, 16, 32, 64, 64, 64]

    times = []
    for _ in range(3):
        image = np.random.randint(0, 256, sizes[i])
        height, width = image.shape[:2]
        p = np.stack((
            np.random.randint(0, height, size=num_pts[i]),
            np.random.randint(0, width, size=num_pts[i]),
        ), axis=1)
        q = p + np.random.randint(-20, 20, size=p.shape)

        start = time.time()
        _ = benchmark_numpy(image, p, q)
        elapse = time.time() - start
        times.append(elapse)
    print("Time (numpy):", sum(times) / len(times))

    times = []
    for _ in range(3):
        image = torch.randint(0, 256, sizes[i]).to(device)
        height, width = image.shape[:2]
        p = torch.stack((
            torch.randint(0, height, size=(num_pts[i],)),
            torch.randint(0, width, size=(num_pts[i],)),
        ), dim=1).to(device)
        q = p + torch.randint(-20, 20, size=p.shape).to(device)

        start = time.time()
        _ = benchmark_torch(image, p, q)
        elapse = time.time() - start
        times.append(elapse)
    print("Time (torch):", sum(times) / len(times))


if __name__ == "__main__":
    #demo()
    # demo_torch()
    demo2()
    # demo3()

    # run_benckmark(i=0)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Image deformation using moving least squares.

    * Affine deformation
    * Similarity deformation
    * Rigid deformation

For more details please refer to the Chinese documentation: 
    
    ./doc/Image Deformation.pdf

or the original paper: 
    
    Image deformation using moving least squares
    Schaefer, Mcphail, Warren. 

Note:
    In the original paper, the author missed the weight w_j in formular (5).
    In addition, all the formulars in section 2.1 miss the w_j. 
    And I have corrected this point in my documentation.

@author: Jian-Wei ZHANG
@email: zjw.cs@zju.edu.cn
@date: 2022/01/12: PyTorch implementation
"""

import torch

from interp_torch import interp


def mls_affine_deformation(vy, vx, p, q, alpha=1.0, eps=1e-8):
    """
    Affine deformation

    Parameters
    ----------
    vy, vx: torch.Tensor
        coordinate grid, generated by torch.meshgrid(gridX, gridY)
    p: torch.Tensor
        an array with size [n, 2], original control points, in (y, x) formats
    q: torch.Tensor
        an array with size [n, 2], final control points, in (y, x) formats
    alpha: float
        parameter used by weights
    eps: float
        epsilon
    
    Return
    ------
        A deformed image.
    """

    device = q.device
    # Change (x, y) to (row, col)
    q = q.short()
    p = p.short()

    # Exchange p and q and hence we transform destination pixels to the corresponding source pixels.
    p, q = q, p

    grow = vx.shape[0]  # grid rows
    gcol = vx.shape[1]  # grid cols
    ctrls = p.shape[0]  # control points

    # Precompute
    reshaped_p = p.reshape(ctrls, 2, 1, 1)                                              # [ctrls, 2, 1, 1]
    reshaped_v = torch.cat((vx.reshape(1, grow, gcol), vy.reshape(1, grow, gcol)), dim=0)      # [2, grow, gcol]

    w = 1.0 / (torch.sum((reshaped_p - reshaped_v).float() ** 2, dim=1) + eps) ** alpha    # [ctrls, grow, gcol]
    w /= torch.sum(w, dim=0, keepdim=True)                                               # [ctrls, grow, gcol]

    pstar = torch.zeros((2, grow, gcol), dtype=torch.float32).to(device)
    for i in range(ctrls):
        pstar += w[i] * reshaped_p[i]                                                   # [2, grow, gcol]

    phat = reshaped_p - pstar                                                           # [ctrls, 2, grow, gcol]
    phat = phat.reshape(ctrls, 2, 1, grow, gcol)                                        # [ctrls, 2, 1, grow, gcol]
    phat1 = phat.reshape(ctrls, 1, 2, grow, gcol)                                       # [ctrls, 1, 2, grow, gcol]
    reshaped_w = w.reshape(ctrls, 1, 1, grow, gcol)                                     # [ctrls, 1, 1, grow, gcol]
    pTwp = torch.zeros((2, 2, grow, gcol), dtype=torch.float32).to(device)
    for i in range(ctrls):
        pTwp += phat[i] * reshaped_w[i] * phat1[i]
    del phat1

    try:
        inv_pTwp = torch.inverse(pTwp.permute(2, 3, 0, 1))                            # [grow, gcol, 2, 2]
        flag = False
    except RuntimeError as error:
        # print(error)
        flag = True             
        det = torch.det(pTwp.permute(2, 3, 0, 1))                                 # [grow, gcol]
        det[det < 1e-8] = float("Inf")
        reshaped_det = det.reshape(1, 1, grow, gcol)                                    # [1, 1, grow, gcol]
        adjoint = pTwp[[[1, 0], [1, 0]], [[1, 1], [0, 0]], :, :]                        # [2, 2, grow, gcol]
        adjoint[[0, 1], [1, 0], :, :] = -adjoint[[0, 1], [1, 0], :, :]                  # [2, 2, grow, gcol]
        inv_pTwp = (adjoint / reshaped_det).permute(2, 3, 0, 1)                       # [grow, gcol, 2, 2]
    
    mul_left = reshaped_v - pstar                                                       # [2, grow, gcol]
    reshaped_mul_left = mul_left.reshape(1, 2, grow, gcol).permute(2, 3, 0, 1)        # [grow, gcol, 1, 2]
    mul_right = torch.mul(reshaped_w, phat, out=phat)                                 # [ctrls, 2, 1, grow, gcol]
    reshaped_mul_right = mul_right.permute(0, 3, 4, 1, 2)                             # [ctrls, grow, gcol, 2, 1]
    out_A = mul_right.reshape(2, ctrls, grow, gcol, 1, 1)[0]                            # [ctrls, grow, gcol, 1, 1]
    A = torch.matmul(torch.matmul(reshaped_mul_left, inv_pTwp), reshaped_mul_right, out=out_A)    # [ctrls, grow, gcol, 1, 1]
    A = A.reshape(ctrls, 1, grow, gcol)                                                 # [ctrls, 1, grow, gcol]
    del mul_right, reshaped_mul_right, phat

    # Calculate q
    reshaped_q = q.reshape((ctrls, 2, 1, 1))                                            # [ctrls, 2, 1, 1]
    qstar = torch.zeros((2, grow, gcol), dtype=torch.float32).to(device)
    for i in range(ctrls):
        qstar += w[i] * reshaped_q[i]                                                   # [2, grow, gcol]
    del w, reshaped_w

    # Get final image transfomer -- 3-D array
    transformers = torch.zeros((2, grow, gcol), dtype=torch.float32).to(device)
    for i in range(ctrls):
        transformers += A[i] * (reshaped_q[i] - qstar)
    transformers += qstar
    del A

    # Correct the points where pTwp is singular
    if flag:
        blidx = det == float('Inf')    # bool index
        transformers[0][blidx] = vx[blidx] + qstar[0][blidx] - pstar[0][blidx]
        transformers[1][blidx] = vy[blidx] + qstar[1][blidx] - pstar[1][blidx]

    # Removed the points outside the border
    transformers[transformers < 0] = 0
    transformers[0][transformers[0] > grow - 1] = 0
    transformers[1][transformers[1] > gcol - 1] = 0

    return transformers.long()


def mls_similarity_deformation(vy, vx, p, q, alpha=1.0, eps=1e-8):
    """ Similarity deformation
    
    Parameters
    ----------
    vx, vy: torch.Tensor
        coordinate grid, generated by torch.meshgrid(gridX, gridY)
    p: torch.Tensor
        an array with size [n, 2], original control points, in (y, x) formats
    q: torch.Tensor
        an array with size [n, 2], final control points, in (y, x) formats
    alpha: float
        parameter used by weights
    eps: float
        epsilon
    
    Return
    ------
        A deformed image.
    """
    device = q.device
    q = q.short()
    p = p.short()

    # Exchange p and q and hence we transform destination pixels to the corresponding source pixels.
    p, q = q, p

    grow = vx.shape[0]  # grid rows
    gcol = vx.shape[1]  # grid cols
    ctrls = p.shape[0]  # control points

    # Compute
    reshaped_p = p.reshape(ctrls, 2, 1, 1)                                              # [ctrls, 2, 1, 1]
    reshaped_v = torch.cat((vx.reshape(1, grow, gcol), vy.reshape(1, grow, gcol)), dim=0)      # [2, grow, gcol]
    
    w = 1.0 / (torch.sum((reshaped_p - reshaped_v).float() ** 2, dim=1) + eps) ** alpha    # [ctrls, grow, gcol]
    w /= torch.sum(w, dim=0, keepdim=True)                                               # [ctrls, grow, gcol]

    pstar = torch.zeros((2, grow, gcol), dtype=torch.float32).to(device)
    for i in range(ctrls):
        pstar += w[i] * reshaped_p[i]                                                   # [2, grow, gcol]

    phat = reshaped_p - pstar                                                           # [ctrls, 2, grow, gcol]
    reshaped_phat = phat.reshape(ctrls, 1, 2, grow, gcol)                               # [ctrls, 1, 2, grow, gcol]
    reshaped_w = w.reshape(ctrls, 1, 1, grow, gcol)                                     # [ctrls, 1, 1, grow, gcol]
    
    mu = torch.zeros((grow, gcol), dtype=torch.float32).to(device)
    for i in range(ctrls):
        mu += w[i] * (phat[i] ** 2).sum(0)
    reshaped_mu = mu.reshape(1, grow, gcol)                                             # [1, grow, gcol]

    vpstar = reshaped_v - pstar                                                         # [2, grow, gcol]
    reshaped_vpstar = vpstar.reshape(2, 1, grow, gcol)                                  # [2, 1, grow, gcol]
    neg_vpstar_verti = vpstar[[1, 0],...]                                               # [2, grow, gcol]
    neg_vpstar_verti[1,...] = -neg_vpstar_verti[1,...]                                  
    reshaped_neg_vpstar_verti = neg_vpstar_verti.reshape(2, 1, grow, gcol)              # [2, 1, grow, gcol]
    mul_right = torch.cat((reshaped_vpstar, reshaped_neg_vpstar_verti), dim=1)    # [2, 2, grow, gcol]

    # Calculate q
    reshaped_q = q.reshape((ctrls, 2, 1, 1))                                            # [ctrls, 2, 1, 1]
    qstar = torch.zeros((2, grow, gcol), dtype=torch.float32).to(device)
    for i in range(ctrls):
        qstar += w[i] * reshaped_q[i]                                                   # [2, grow, gcol]
    
    # Get final image transfomer -- 3-D array
    temp = torch.zeros((grow, gcol, 2), dtype=torch.float32).to(device)
    for i in range(ctrls):
        neg_phat_verti = phat[i, [1, 0]]                                                # [2, grow, gcol]
        neg_phat_verti[1] = -neg_phat_verti[1]
        reshaped_neg_phat_verti = neg_phat_verti.reshape(1, 2, grow, gcol)              # [1, 2, grow, gcol]
        mul_left = torch.cat((reshaped_phat[i], reshaped_neg_phat_verti), dim=0)  # [2, 2, grow, gcol]
        
        A = torch.matmul((reshaped_w[i] * mul_left).permute(2, 3, 0, 1), 
                      mul_right.permute(2, 3, 0, 1))                                  # [grow, gcol, 2, 2]
    
        qhat = reshaped_q[i] - qstar                                                    # [2, grow, gcol]
        reshaped_qhat = qhat.reshape(1, 2, grow, gcol).permute(2, 3, 0, 1)            # [grow, gcol, 1, 2]

        # Get final image transfomer -- 3-D array
        temp += torch.matmul(reshaped_qhat, A).reshape(grow, gcol, 2)                      # [grow, gcol, 2]

    transformers = temp.permute(2, 0, 1) / reshaped_mu  + qstar                       # [2, grow, gcol]

    # Removed the points outside the border
    transformers[transformers < 0] = 0
    transformers[0][transformers[0] > grow - 1] = 0
    transformers[1][transformers[1] > gcol - 1] = 0

    return transformers.long()


def mls_rigid_deformation(vy, vx, p, q, alpha=1.0, eps=1e-8):
    """ Rigid deformation
    
    Parameters
    ----------
    vx, vy: torch.Tensor
        coordinate grid, generated by torch.meshgrid(gridX, gridY)
    p: torch.Tensor
        an array with size [n, 2], original control points, in (y, x) formats
    q: torch.Tensor
        an array with size [n, 2], final control points, in (y, x) formats
    alpha: float
        parameter used by weights
    eps: float
        epsilon
    
    Return
    ------
        A deformed image.
    """
    device = q.device
    q = q.short()
    p = p.short()

    # Exchange p and q and hence we transform destination pixels to the corresponding source pixels.
    p, q = q, p

    grow = vx.shape[0]  # grid rows
    gcol = vx.shape[1]  # grid cols
    ctrls = p.shape[0]  # control points

    # Compute
    reshaped_p = p.reshape(ctrls, 2, 1, 1)                                              # [ctrls, 2, 1, 1]
    reshaped_v = torch.cat((vx.reshape(1, grow, gcol), vy.reshape(1, grow, gcol)), dim=0)      # [2, grow, gcol]
    
    w = 1.0 / (torch.sum((reshaped_p - reshaped_v).float() ** 2, dim=1) + eps) ** alpha    # [ctrls, grow, gcol]
    w /= torch.sum(w, dim=0, keepdim=True)                                               # [ctrls, grow, gcol]
    
    pstar = torch.zeros((2, grow, gcol), dtype=torch.float32).to(device)
    for i in range(ctrls):
        pstar += w[i] * reshaped_p[i]                                                   # [2, grow, gcol]

    vpstar = reshaped_v - pstar                                                         # [2, grow, gcol]
    reshaped_vpstar = vpstar.reshape(2, 1, grow, gcol)                                  # [2, 1, grow, gcol]
    neg_vpstar_verti = vpstar[[1, 0],...]                                               # [2, grow, gcol]
    neg_vpstar_verti[1,...] = -neg_vpstar_verti[1,...]                                  
    reshaped_neg_vpstar_verti = neg_vpstar_verti.reshape(2, 1, grow, gcol)              # [2, 1, grow, gcol]
    mul_right = torch.cat((reshaped_vpstar, reshaped_neg_vpstar_verti), dim=1)    # [2, 2, grow, gcol]
    reshaped_mul_right = mul_right.reshape(2, 2, grow, gcol)                            # [2, 2, grow, gcol]

    # Calculate q
    reshaped_q = q.reshape((ctrls, 2, 1, 1))                                            # [ctrls, 2, 1, 1]
    qstar = torch.zeros((2, grow, gcol), dtype=torch.float32).to(device)
    for i in range(ctrls):
        qstar += w[i] * reshaped_q[i]                                                   # [2, grow, gcol]
    
    temp = torch.zeros((grow, gcol, 2), dtype=torch.float32).to(device)
    for i in range(ctrls):
        phat = reshaped_p[i] - pstar                                                    # [2, grow, gcol]
        reshaped_phat = phat.reshape(1, 2, grow, gcol)                                  # [1, 2, grow, gcol]
        reshaped_w = w[i].reshape(1, 1, grow, gcol)                                     # [1, 1, grow, gcol]
        neg_phat_verti = phat[[1, 0]]                                                   # [2, grow, gcol]
        neg_phat_verti[1] = -neg_phat_verti[1]
        reshaped_neg_phat_verti = neg_phat_verti.reshape(1, 2, grow, gcol)              # [1, 2, grow, gcol]
        mul_left = torch.cat((reshaped_phat, reshaped_neg_phat_verti), dim=0)     # [2, 2, grow, gcol]
        
        A = torch.matmul((reshaped_w * mul_left).permute(2, 3, 0, 1), 
                        reshaped_mul_right.permute(2, 3, 0, 1))                       # [grow, gcol, 2, 2]

        qhat = reshaped_q[i] - qstar                                                    # [2, grow, gcol]
        reshaped_qhat = qhat.reshape(1, 2, grow, gcol).permute(2, 3, 0, 1)            # [grow, gcol, 1, 2]

        # Get final image transfomer -- 3-D array
        temp += torch.matmul(reshaped_qhat, A).reshape(grow, gcol, 2)                      # [grow, gcol, 2]

    temp = temp.permute(2, 0, 1)                                                      # [2, grow, gcol]
    normed_temp = torch.norm(temp, dim=0, keepdim=True)                           # [1, grow, gcol]
    normed_vpstar = torch.norm(vpstar, dim=0, keepdim=True)                       # [1, grow, gcol]
    transformers = temp / normed_temp * normed_vpstar  + qstar                          # [2, grow, gcol]
    nan_mask = normed_temp[0] == 0

    # Replace nan values by interpolated values
    nan_mask_flat = torch.nonzero(nan_mask.view(-1), as_tuple=True)[0]
    nan_mask_anti_flat = torch.nonzero(~nan_mask.view(-1), as_tuple=True)[0]
    transformers[0][nan_mask] = interp(nan_mask_flat, nan_mask_anti_flat, transformers[0][~nan_mask])
    transformers[1][nan_mask] = interp(nan_mask_flat, nan_mask_anti_flat, transformers[1][~nan_mask])

    # Remove the points outside the border
    transformers[transformers < 0] = 0
    transformers[0][transformers[0] > grow - 1] = 0
    transformers[1][transformers[1] > gcol - 1] = 0
    
    return transformers.long()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


_author__ = "Patrice Guyot"
__email__ = "patrice.guyot@irit.fr"
__version__ = "1.1.0"

"""
***********************************************************************
Name            : audio_processing.py
Description     : The module implements a function to read audio file.
Author          : Patrice Guyot.
***********************************************************************
"""



from scipy.io.wavfile import read as wavread
import logging
from os import remove, sep
import subprocess



def audio_read(audioFilePath, formatsox=False):
    """

    Read an audio file (from scipy.io.wavfile)

    A conversation of the aufio file can be processed using SOX (by default set at False).

    :param audioFilePath: audio file name (with eventually the full path)
    :type audioFilePath: str
    :param formatsox: if set at True, us SOX to convert audio file in "wav file, 1 channel, 16 kHz, 16bits". Default: False
    :type audioFilePath: str

    :returns:
        * sr: sampling rate of the signal
        * sig: list of values of the signal
    :rtype: tuple

    """
    logging.info('Reading of the audio file : ' + audioFilePath)
    if formatsox:
        tmpFile = "tmp.wav"
        logging.info('\t- Conversion en wav 16k with SOX.')
        cmd = "sox " + audioFilePath + " -c 1 -r 16k -b 16 -G " + tmpFile + " rate -m"
        subprocess.check_call(cmd)
        [sr, sig] = wavread(tmpFile)
        remove(tmpFile)
    else:
        [sr, sig] = wavread(audioFilePath)

    return sr, sig




#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
CLEESE toolbox
v1.0: mar 2018, J.J. Burred <jjburred@jjburred.com> for IRCAM/CNRS
v2.0: jan 2022, Lara Kermarec <lara.git@kermarec.bzh> for CNRS

Main functions for CLEESE's phase vocoder engine
'''

import os
import numpy as np
import scipy.io.wavfile as wav
import scipy.signal as sig
import time
import shutil

from .bpf import (
        createBPFtimeVec,
        createBPFfreqs,
        createBPF,
        create_BPF_header)
from .audio_engine import (
        stft,
        istft,
        istft_resamp,
        phaseVocoder_varHop,)
from ..engine import Engine
from ...cleese import log, load_config

from .utils import load_file, wav_read, wav_write, extract_pitch, extract_spectral_env


class PhaseVocoder(Engine):

    @staticmethod
    def generate_stimuli(soundData, config, BPF=None, sample_rate=None,
                         sample_format=None,
                         timeVec=None):
        PhaseVocoder.process(
                soundData, config, BPF=BPF, sample_rate=sample_rate,
                sample_format=sample_format, timeVec=timeVec,
                file_output=True)

    @staticmethod
    def process(soundData, config, BPF=None, sample_rate=None,
                sample_format=None, timeVec=None,
                file_output=False):

        doCreateBPF = False
        if BPF is None:
            doCreateBPF = True

        try:
            BPF_EXT = config["main"]["param_ext"]
        except KeyError as e:
            BPF_EXT = '.txt' # default bpf extension
  

        if not sample_rate:
            log("ERROR: missing sample rate")
            return

        if file_output:
            if not sample_rate or not sample_format:
                log("ERROR: missing sample format")
                return

        sr = sample_rate
        sampleFormat = sample_format

        waveIn = soundData
        numFiles = 1

        if len(waveIn.shape) == 2:
            log('WARN: stereo file detected. Reading only left channel.')
            waveIn = np.ravel(waveIn[:, 0])

        config["main"]['inSamples'] = len(waveIn)
        config["main"]['sr'] = sr

        if doCreateBPF and file_output:

            numFiles = config["main"]['numFiles']

        else:

            numFiles = 1
            if np.isscalar(BPF):
                BPF = np.array([[0., float(BPF)]])

        currOutFile = []
        for t in range(0, len(config["main"]['transf'])):

            tr = config["main"]['transf'][t]

            if config["main"]['chain']:
                if t == 0:
                    currTrString = tr
                else:
                    currTrString = currTrString+'_'+tr
            else:
                currTrString = tr

            if doCreateBPF:

                if file_output:
                    config["main"]['currOutPath'] = os.path.join(config['main']['expBaseDir'], currTrString)

                # create BPF time vector
                duration = config["main"]['inSamples']/float(sr)
                BPFtime, numPoints, endOnTrans = createBPFtimeVec(
                        duration, config[tr], timeVec=timeVec)

            else:
                config["main"]['currOutPath'] = config["main"]['outPath']

            # create output folder
            if file_output:
                if not os.path.exists(config["main"]['currOutPath']):
                    os.makedirs(config["main"]['currOutPath'])
                path, inFileNoExt = os.path.split(config["main"]["filename"])
                inFileNoExt = os.path.splitext(inFileNoExt)[0]

            # create frequency bands for random EQ
            eqFreqVec = None
            if tr == 'eq':
                eqFreqVec = createBPFfreqs(config)

            for i in range(0, numFiles):

                log(currTrString+' variation '+str(i+1)+'/'+str(numFiles))
                currFileNo = "%08u" % (i+1)

                if config["main"]['chain'] and t > 0:
                    if file_output:
                        waveIn, sr, sampleFormat = PhaseVocoder.wav_read(
                                currOutFile[i])
                    else:
                        waveIn = waveOut
                    config["main"]['inSamples'] = len(waveIn)
                    config["main"]['sr'] = sr
                    duration = len(waveIn) / float(sr)
                    BPFtime, numPoints, endOnTrans = createBPFtimeVec(
                            duration, config[tr], timeVec=timeVec)

                # generate random BPF
                if doCreateBPF:
                    BPF = createBPF(tr, config, BPFtime, numPoints,
                                    endOnTrans, eqFreqVec)

                # export BPF as text file
                if file_output:
                    currBPFfile = os.path.join(
                            config["main"]['currOutPath'],
                            inFileNoExt+'.'+currFileNo+'.'+currTrString+BPF_EXT)
                    BPF_header = create_BPF_header(tr, config)

                    np.savetxt(currBPFfile, BPF, '%.8f', delimiter = ',', header = BPF_header, comments='')

                    if t == 0:
                        currOutFile.append(os.path.join(
                                config["main"]['currOutPath'],
                                inFileNoExt+'.'+currFileNo+'.'+currTrString+'.wav'))
                    else:
                        currOutFile[i] = os.path.join(
                                config["main"]['currOutPath'],
                                inFileNoExt+'.'+currFileNo+'.'+currTrString+'.wav')

                if tr in ['stretch', 'pitch']:

                    if tr == 'stretch':
                        doPitchShift = False
                    elif tr == 'pitch':
                        doPitchShift = True

                    # call processing with phase vocoder
                    waveOut = processWithPV(waveIn=waveIn,
                                            config=config,
                                            BPF=BPF,
                                            doPitchShift=doPitchShift)

                    # remove trailing zero-pad
                    if tr == 'pitch':
                        waveOut = np.delete(
                                waveOut,
                                range(
                                    config["main"]['inSamples'],
                                    len(waveOut)))

                elif tr == 'eq':

                    waveOut = processWithSTFT(
                            waveIn=waveIn, config=config, BPF=BPF)

                    # remove trailing zero-pad
                    waveOut = np.delete(
                            waveOut,
                            range(config["main"]['inSamples'], len(waveOut)))

                elif tr == 'gain':

                    if numPoints == 1:
                        waveOut = waveIn * BPF[:, 1]
                    else:
                        gainVec = np.interp(np.linspace(
                                0, duration, config["main"]['inSamples']),
                                BPF[:, 0], BPF[:, 1])
                        waveOut = waveIn * gainVec

                # if gain > 1.0, normalize to avoid clipping
                if np.max(np.abs(waveOut)) >= 1.0:
                    waveOut = waveOut/np.max(np.abs(waveOut))*0.999
                
                if file_output:            
                    PhaseVocoder.wav_write(waveOut, file_name=currOutFile[i],
                                          sr=sr, sample_format=sampleFormat)

        if not file_output:
            return waveOut, BPF

    @staticmethod
    def name():
        return "phase_vocoder"

    @staticmethod
    def load_file(file_name): 
        return load_file(file_name)

    @staticmethod
    def wav_read(file_name):
        return wav_read(file_name)

    @staticmethod
    def wav_write(wave_out, file_name, sr, sample_format='int16'):
        return wav_write(wave_out, file_name, sr, sample_format)

    @staticmethod
    def extract_pitch(x, sr, win=.02, bounds=[70,400], interpolate=True):
        return extract_pitch(x, sr, win, bounds, interpolate)

    @staticmethod
    def extract_spectral_env(x, sr, lpc_order=50, pe_thresh=1000, freq_limit = 20000):
        return extract_spectral_env(x, sr, lpc_order, pe_thresh, freq_limit)


    @staticmethod
    def create_BPF(trans, config_file, time_points, num_points, end_on_transition, eq_freqs=None):
        config = load_config(config_file)
        if config is None:
            return
        return createBPF(trans, config, time_points, num_points, end_on_transition, eq_freqs)


def processWithSTFT(waveIn, config, BPF):

    sr = config["main"]['sr']
    inSamples = config["main"]['inSamples']

    winLen = config["analysis"]["window"]["len"]        # in seconds
    n_fft = int(2**np.ceil(np.log2(winLen*sr)))   # next pow 2
    overlapFactor = config["analysis"]["oversampling"]
    synHop = n_fft//overlapFactor

    win = np.hamming(n_fft)

    numFrames = int(round(inSamples/synHop)) + overlapFactor
    anaHopVec = np.ones(numFrames, dtype=int) * synHop

    stftMat = stft(waveIn, win, n_fft, anaHopVec)
    stftMat = np.squeeze(stftMat)
    stftMat = stftMat[0:n_fft//2+1, :]

    # interpolate BPF matrix to generate filter mask
    numSeg = BPF.shape[0]
    numBands = (BPF.shape[1]-2)//2

    freqMat = BPF[:, 2::2]
    amplMat = BPF[:, 3::2]
    timeVec = BPF[:, 0]

    # interpolate in frequency
    interpBPF = np.zeros((n_fft//2+1, numSeg))
    for i in range(0, numSeg):
        currBinVec = freqMat[i, :]/sr*2 * (n_fft//2+1)
        # piecewise-linear spectral envelope
        interpBPF[:, i] = np.interp(np.arange(0, n_fft//2+1),
                                    currBinVec, amplMat[i, :],
                                    left=0, right=0)

    # interpolate in time
    filterMat = np.zeros(stftMat.shape)
    if numSeg == 1:
        filterMat = np.tile(interpBPF, (1, stftMat.shape[1]))
    else:
        frameVec = timeVec*sr/synHop
        for i in range(0, n_fft//2+1):
            filterMat[i, :] = np.interp(np.arange(0, numFrames),
                                        frameVec, interpBPF[i, :])

    stftMat *= np.power(10,filterMat/20)

    waveOut = istft(stftMat, win, n_fft, synHop)

    return waveOut


def processWithPV(waveIn, config, BPF, doPitchShift=False):

    sr = config["main"]["sr"]
    inSamples = config["main"]["inSamples"]

    if not doPitchShift:

        # static stretch
        if BPF.size == 2:
            rateVec = 1/BPF[0, 1]  # >1: dilation (stretching), <1: compression
            numSeg = 1
        else:
            sampleVec = np.rint(BPF[:, 0]*sr)   # in samples
            rateVec = BPF[:, 1]
            numSeg = BPF.shape[0]

    else:

        bins_per_octave = 12

        # static pitch shift
        if BPF.size == 2:
            # in libRosa, pitch shifting factors are in semitones.
            # In CLEESE, in cents.
            rateVec = 2.0 ** (-float(BPF[0, 1]/100.0) / bins_per_octave)
            numSeg = 1
        else:
            sampleVec = np.rint(BPF[:, 0]*sr)   # in samples
            rateVec = 2.0 ** (BPF[:, 1] / float(100*bins_per_octave))

            numSeg = BPF.shape[0]

    winLen = config["analysis"]["window"]["len"]  # in seconds
    n_fft = int(2**np.ceil(np.log2(winLen*sr)))   # next pow 2
    overlapFactor = config["analysis"]["oversampling"]
    synHop = n_fft//overlapFactor

    if numSeg == 1:

        anaHop = int(round(synHop*rateVec))
        numFrames = int(round(inSamples/anaHop)) + overlapFactor
        anaHopVec = np.ones(numFrames, dtype=int) * anaHop

    else:

        pos = 0
        currHop = int(round(synHop/rateVec[0]))
        anaHopVec = currHop
        allRatesVec = rateVec[0]

        while pos <= inSamples:
            pos += currHop
            sampledBPF = np.interp(pos, sampleVec, rateVec)
            currHop = int(round(synHop/sampledBPF))
            anaHopVec = np.append(anaHopVec, currHop)
            allRatesVec = np.append(allRatesVec, sampledBPF)

    win = np.hamming(n_fft)

    stftMat = stft(waveIn, win, n_fft, anaHopVec)

    stftMat = np.squeeze(stftMat)
    stftMat = stftMat[0:n_fft//2+1, :]

    phase_locking = 1

    stft_stretch = phaseVocoder_varHop(stftMat,
                                       anaHopVec=anaHopVec,
                                       synHop=synHop,
                                       phase_locking=phase_locking)

    if doPitchShift == 0:

        waveOut = istft(stft_stretch, win, n_fft, synHop)

    else:
        if numSeg == 1:

            waveOut = istft(stft_stretch, win, n_fft, synHop)
            n_samples = int(np.ceil(len(waveOut) * rateVec))
            waveOut = sig.resample_poly(waveOut, n_samples, len(waveOut))

        else:

            waveOut = istft_resamp(stft_stretch, win, n_fft,
                                   synHop, allRatesVec, inSamples)

    return waveOut
